import os
import tkinter as tk

from mmg_toolbox.utils.env_functions import get_notebook_directory, open_terminal, get_scan_number
from mmg_toolbox.tkguis.misc.config import get_config, C
from mmg_toolbox.tkguis.misc.functions import topmenu
from mmg_toolbox.tkguis.misc.styles import RootWithStyle, create_root
from mmg_toolbox.tkguis.misc.jupyter import launch_jupyter_notebook, terminate_notebooks
from mmg_toolbox.scripts.scripts import create_script, create_notebook, SCRIPTS, NOTEBOOKS


def create_data_viewer(initial_folder: str | None = None,
                       parent: tk.Misc | None = None, config: dict | None = None) -> RootWithStyle:
    """
    Create a Data Viewer showing all scans in an experiment folder
    """
    from ..widgets.nexus_data_viewer import NexusDataViewer
    from .log_viewer import create_gda_terminal_log_viewer
    from .file_browser import create_nexus_file_browser, create_file_browser, create_jupyter_browser
    from .multi_scan_analysis import create_multi_scan_analysis
    from .scans import create_range_selector
    from .python_editor import create_python_editor

    root = create_root(parent=parent, window_title='NeXus Data Viewer')
    config = config or get_config()

    widget = NexusDataViewer(root, initial_folder=initial_folder, config=config)

    def get_filepath():
        filename, folder = widget.selector_widget.get_filepath()
        return folder

    def get_replacements(*filenames):
        return {
            # {{template}}: replacement
            'description': 'an example script',
            'filepaths': ', '.join(f"'{f}'" for f in filenames),
            'title': f"Example Script: {os.path.basename(filenames[0])}",
            'x-axis': widget.plot_widget.axes_x.get(),
            'y-axis': widget.plot_widget.axes_y.get(),
        }

    def create_script_template(template='example'):
        filename, folder = widget.selector_widget.get_filepath()
        new_file = folder + '/processing/example_script.py'
        create_script(new_file, template, **get_replacements(filename))
        create_python_editor(open(new_file).read(), root, config),

    def create_notebook_template(template='example'):
        filename, folder = widget.selector_widget.get_filepath()
        new_file = folder + '/processing/example.ipynb'
        create_notebook(new_file, template, **get_replacements(filename))
        launch_jupyter_notebook('notebook', file=new_file)

    def start_multi_scan_plot():
        filename, folder = widget.selector_widget.get_filepath()
        filenames = widget.selector_widget.get_multi_filepath()
        scan_numbers = [get_scan_number(f) for f in filenames]
        create_multi_scan_analysis(root, config, exp_directory=folder, scan_numbers=scan_numbers)

    menu = {
        'File': {
            'New Data Viewer': lambda: create_data_viewer(parent=root, config=config),
            'Add Folder': widget.selector_widget.browse_folder,
            'File Browser': lambda: create_file_browser(root, config.get(C.default_directory, None)),
            'NeXus File Browser': lambda: create_nexus_file_browser(root, config.get(C.default_directory, None)),
            'Jupyter Browser': lambda: create_jupyter_browser(root, get_notebook_directory(get_filepath())),
            'Range selector': lambda: create_range_selector(initial_folder, root, config),
            'Log viewer': lambda: create_gda_terminal_log_viewer(get_filepath(), root)
        },
        'Processing': {
            'Multi-Scan': start_multi_scan_plot,
            'Script Editor': lambda: create_python_editor(None, root, config),
            'Open a terminal': lambda: open_terminal(f"cd {get_filepath()}"),
            'Start Jupyter (processing)': lambda: launch_jupyter_notebook('notebook', get_filepath() + '/processing'),
            'Start Jupyter (notebooks)': lambda: launch_jupyter_notebook('notebook', get_filepath() + '/processed/notebooks'),
            'Stop Jupyter servers': terminate_notebooks,
            'Scripts:': {name: lambda n=name: create_script_template(n) for name in SCRIPTS},
            'Notebooks:': {name: lambda n=name: create_notebook_template(n) for name in NOTEBOOKS},
        }
    }
    menu.update(widget.image_widget.options_menu())

    topmenu(root, menu, add_themes=True, add_about=True, config=config)

    root.update()

    if parent is None:
        root.mainloop()
    return root



"""
mmg_toolbox tests
Test utilities
"""
import numpy as np

from mmg_toolbox.utils import xray_utils, rotations, misc_functions, units


def test_wavelength():
    assert abs(xray_utils.photon_energy(1.0) - 12.398) < 0.01
    assert abs(xray_utils.photon_wavelength(8) - 1.55) < 0.01
    assert abs(xray_utils.wavevector(1.0) - 6.283) < 0.01


def test_rotation():
    rotation = rotations.rotation_t_matrix(30, (1, 1, 0))
    translation = rotations.translation_t_matrix(5, (0, 0, 1))
    total = translation @ rotation
    vec = (1, 1, 1)
    new_vec1 = rotations.transform_by_t_matrix(vec, rotation)
    new_vec2 = rotations.transform_by_t_matrix(new_vec1, translation)
    new_vec3 = rotations.transform_by_t_matrix(vec, total)
    assert np.allclose(new_vec2, new_vec3)


def test_units():
    distance = units.unit_converter(10, 'nm', 'A')
    assert distance == 100
    energy = units.unit_converter(8, 'keV', 'eV')
    assert energy == 8000


def test_misc_functions():
    assert misc_functions.stfm(35.25, 0.01) == '35.25 (1)'
    assert misc_functions.stfm(110.25, 5) == '110 (5)'
    assert misc_functions.stfm(1.5632e6,1.53e4) == '1.56(2)E+6'

    # misc_functions.numbers2string()





"""
mmg_toolbox tests
Test tkinter dataviewer
"""

from mmg_toolbox.tkguis import create_title_window, create_data_viewer, create_nexus_viewer, create_nexus_file_browser
from . import only_dls_file_system


def test_file_loader():
    # currently don't have a good way of doing this
    # root.mainloop() calls freeze the test
    assert True




"""
Configuration Options
"""
from __future__ import annotations

import os
import json

from mmg_toolbox.utils.env_functions import TMPDIR, YEAR, get_beamline, get_user, check_file_access
from .beamline_metadata import BEAMLINE_META, META_STRING, META_LABEL
from .matplotlib import FIGURE_SIZE, FIGURE_DPI, IMAGE_SIZE, DEFAULT_COLORMAP


class C:
    """Names used in config object"""
    conf_file = 'config_file'
    default_directory = 'default_directory'
    processing_directory = 'processing_directory'
    notebook_directory = 'notebook_directory'
    recent_data_directories = 'recent_data_directories'
    small_screen_height = 'small_screen_height'
    text_size = 'text_size'
    text_size_small = 'text_size_small'
    plot_size = 'plot_size'
    image_size = 'image_size'
    plot_max_percent = 'plot_max_percent'
    plot_dpi = 'plot_dpi'
    plot_title = 'plot_title'
    normalise_factor = 'normalise_factor'
    replace_names = 'replace_names'
    metadata_string = 'metadata_string'
    metadata_list = 'metadata_list'
    metadata_label = 'metadata_label'
    default_colormap = 'default_colormap'
    beamline = 'beamline'
    roi = 'roi'
    current_dir = 'current_dir'
    current_proc = 'current_proc'
    current_nb = 'current_nb'


# config name (saved in TMPDIR)
USER = get_user()
TMPFILE = f'mmg_config_{USER}.json'
CONFIG_FILE = os.path.join(TMPDIR, TMPFILE)
SMALL_SCREEN_HEIGHT = 800  # pixels, reduce size if screen smaller than this
TEXT_WIDTH = 50  # Determines the width of text areas in DataViewer in characters
TEXT_HEIGHT = 25  # Determines height of text area in Dataviewer in lines
TEXT_HEIGHT_SMALL = 10  # TEXT_HEIGHT when screen is small
MAX_PLOT_SCREEN_PERCENTAGE = (75, 25)  # (wid, height) max plot size as % of screen


META_LIST = {
    # scan number and start_time included by default
    # name: format
    'cmd': '{(cmd|scan_command)}'
}

REPLACE_NAMES = {
    # NEW_NAME: EXPRESSION
    '_t': '(count_time|counttime|t?(1.0))',
}

ROIs: list[tuple[str, str | int, str | int, int, int, str]] = [
    # (name, cen_i, cen_j, wid_i, wid_j, det_name)
]

CONFIG = {
    C.conf_file: CONFIG_FILE,
    C.default_directory: os.path.expanduser('~'),
    C.processing_directory: os.path.expanduser('~'),
    C.notebook_directory: os.path.expanduser('~'),
    C.recent_data_directories: [os.path.expanduser('~')],
    C.small_screen_height: SMALL_SCREEN_HEIGHT,
    C.text_size: (TEXT_WIDTH, TEXT_HEIGHT),
    C.text_size_small: (TEXT_WIDTH, TEXT_HEIGHT_SMALL),
    C.plot_size: FIGURE_SIZE,
    C.image_size: IMAGE_SIZE,
    C.plot_max_percent: MAX_PLOT_SCREEN_PERCENTAGE,
    C.plot_dpi: FIGURE_DPI,
    C.plot_title: '{filename}\n{(cmd|scan_command)}',
    C.normalise_factor: '',
    C.replace_names: REPLACE_NAMES,
    C.roi: ROIs,
    C.metadata_string: META_STRING,
    C.metadata_list: META_LIST,
    C.metadata_label: META_LABEL,
    C.default_colormap: DEFAULT_COLORMAP,
}

BEAMLINE_CONFIG = {
    'i06': {
        C.beamline: 'i06',
        C.default_directory: f"/dls/i06/data/{YEAR}/",
        C.metadata_string: BEAMLINE_META['i06'],
        C.normalise_factor: '',
    },
    'i06-1': {
        C.beamline: 'i06-1',
        C.default_directory: f"/dls/i06-1/data/{YEAR}/",
        C.metadata_string: BEAMLINE_META['i06-1'],
        C.normalise_factor: '',
    },
    'i06-2': {
        C.beamline: 'i06-2',
        C.default_directory: f"/dls/i06-2/data/{YEAR}/",
        C.metadata_string: BEAMLINE_META['i06-2'],
        C.normalise_factor: '',
    },
    'i10': {
        C.beamline: 'i10',
        C.default_directory: f"/dls/i10/data/{YEAR}/",
        C.metadata_string: BEAMLINE_META['i10'],
        C.normalise_factor: '',
    },
    'i10-1': {
        C.beamline: 'i10-1',
        C.default_directory: f"/dls/i10-1/data/{YEAR}/",
        C.metadata_string: BEAMLINE_META['i10-1'],
        C.normalise_factor: '/(mcs16|macr16|mcse16|macj316|mcsh16|macj216)',
    },
    'i16': {
        C.beamline: 'i16',
        C.default_directory: f"/dls/i16/data/{YEAR}/",
        C.normalise_factor: '/Transmission/count_time/(rc/300.)',
        C.metadata_string: BEAMLINE_META['i16'],
        C.roi: [
            ('pilroi1', 'pil3_centre_j', 'pil3_centre_i', 30, 30, 'pil3_100k'),
        ]
    },
    'i21': {
        C.beamline: 'i21',
        C.default_directory: f"/dls/i21/data/{YEAR}/",
        C.metadata_string: BEAMLINE_META['i21'],
        C.normalise_factor: '',
    },
}


def check_config_filename(config_filename: str | None) -> str:
    """Check config filename is writable, raise OSError if not"""
    if config_filename is None:
        config_filename = CONFIG_FILE
    return check_file_access(config_filename)


def load_config(config_filename: str = CONFIG_FILE) -> dict:
    if os.path.isfile(config_filename):
        with open(config_filename, 'r') as f:
            return json.load(f)
    return {}


def default_config(beamline: str | None = None) -> dict:
    config = CONFIG.copy()
    if beamline is None:
        beamline = get_beamline()
    if beamline in BEAMLINE_CONFIG:
        config.update(BEAMLINE_CONFIG[beamline])
    return config


def get_config(config_filename: str | None = None, beamline: str | None = None) -> dict:
    config_filename = check_config_filename(config_filename)
    user_config = load_config(config_filename)
    config = default_config(beamline)
    config.update(user_config)
    return config


def reset_config(config: dict) -> None:
    """Reset config dict in place with default values of beamline"""
    beamline = config.get(C.beamline, None)
    config.clear()
    config.update(default_config(beamline))


def save_config(config: dict):
    config_filename = config.get(C.conf_file, CONFIG_FILE)
    with open(config_filename, 'w') as f:
        json.dump(config, f)
    print('Saved config to {}'.format(config_filename))


def save_config_as(config_filename: str | None = None, **kwargs):
    config = get_config(config_filename)
    config.update(kwargs)
    config[C.conf_file] = config_filename
    save_config(config)




"""
Script & Notebook templates
"""

import os


SCRIPTS = {
    # name: (filename, description)
    'example': ('example_script.py', 'a simple example'),
    'plot multi-line': ('experiment_multiline.py', 'create a multi-line plot'),
    'spectra': ('spectra_script.py', 'normalise spectra and subtract polarisations')
}

NOTEBOOKS = {
    # name: (filename, description)
    'example': ('example_notebook.ipynb', 'a basic example'),
}

TEMPLATE = {
    # {{template}}: replacement
    'description': 'a short description',
    'filepaths': "'file1.nxs', 'file2.nxs', 'file3.nxs'",
    'experiment_dir': 'path/to/dir',
    'scan_numbers': 'range(-10, 0)',
    'title': 'a nice plot',
    'x-axis': 'axes',
    'y-axis': 'signal',
    'value': 'Ta'
}


def create_script(new_script_path: str, template_name: str, **replacements):
    """create script from template"""
    template_file, description = SCRIPTS[template_name]
    template_file = os.path.join(os.path.dirname(__file__), template_file)
    template_changes = TEMPLATE.copy()
    template_changes.update(replacements)
    print(template_file)
    print(template_changes)

    template_string = open(template_file, 'r').read()
    for name, value in template_changes.items():
        param = "{{" + name + "}}"
        print(f"Replacing {template_string.count(param)} instances of {param}")
        template_string = template_string.replace(param, value)

    with open(new_script_path, 'w') as new:
        new.write(template_string)
    print(f"Created {new_script_path}")


def create_notebook(new_notebook_path: str, template_name: str, **replacements):
    """create script from template"""
    template_file, description = NOTEBOOKS[template_name]
    template_file = os.path.join(os.path.dirname(__file__), template_file)
    template_changes = TEMPLATE.copy()
    template_changes.update(replacements)

    template_string = open(template_file, 'r').read()
    for name, value in template_changes.items():
        param = "{{" + name + "}}"
        print(f"Replacing {template_string.count(param)} instances of {param}")
        template_string = template_string.replace(param, value)

    with open(new_notebook_path, 'w') as new:
        new.write(template_string)
    print(f"Created {new_notebook_path}")




import os

from hdfmap import load_hdf, create_nexus_map

from mmg_toolbox.nexus.nexus_scan import NexusDataHolder, NexusScan
from mmg_toolbox.utils.file_functions import get_scan_number, replace_scan_number


def read_nexus_file(filename: str, flatten_scannables: bool = True) -> NexusDataHolder:
    """
    Read Nexus file as DataHolder
    """
    return NexusDataHolder(filename, flatten_scannables=flatten_scannables)


def read_nexus_files(*filenames: str) -> list[NexusScan]:
    """
    Read Nexus files as NexusScan
    """
    hdf_map = create_nexus_map(filenames[0])
    return [NexusScan(f, hdf_map) for f in filenames]


def find_matching_scans(filename: str, match_field: str = 'scan_command',
                        search_scans_before: int = 10, search_scans_after: int | None = None) -> list[str]:
    """
    Find scans with scan numbers close to the current file with matching scan command

    :param filename: nexus file to start at (must include scan number in filename)
    :param match_field: nexus field to compare between scan files
    :param search_scans_before: number of scans before current scan to look for
    :param search_scans_after: number of scans after current scan to look for (None==before)
    :returns: list of scan files that exist and have matching field values
    """
    nexus_map = create_nexus_map(filename)
    field_value = nexus_map.eval(nexus_map.load_hdf(), match_field)
    scanno = get_scan_number(filename)
    if search_scans_after is None:
        search_scans_after = search_scans_before
    matching_files = []
    for scn in range(scanno - search_scans_before, scanno + search_scans_after):
        new_filename = replace_scan_number(filename, scn)
        if os.path.isfile(new_filename):
            new_field_value = nexus_map.eval(load_hdf(new_filename), match_field)
            if field_value == new_field_value:
                matching_files.append(new_filename)
    return matching_files




"""
X-ray scattering utility funcitons
"""

import numpy as np

# Constants
class Const:
    pi = np.pi  # mmmm tasty Pi
    e = 1.6021733E-19  # C  electron charge
    h = 6.62606868E-34  # Js  Plank consant
    c = 299792458  # m/s   Speed of light
    u0 = 4 * pi * 1e-7  # H m-1 Magnetic permeability of free space
    me = 9.109e-31  # kg Electron rest mass
    mn = 1.6749e-27 # kg Neutron rest mass
    Na = 6.022e23  # Avagadro's No
    A = 1e-10  # m Angstrom
    r0 = 2.8179403227e-15  # m classical electron radius = e^2/(4pi*e0*me*c^2)
    Cu = 8.048  # Cu-Ka emission energy, keV
    Mo = 17.4808  # Mo-Ka emission energy, keV


def photon_wavelength(energy_kev):
    """
    Converts energy in keV to wavelength in A
     wavelength_a = photon_wavelength(energy_kev)
     lambda [A] = h*c/E = 12.3984 / E [keV]
    """

    # Electron Volts:
    E = 1000 * energy_kev * Const.e

    # SI: E = hc/lambda
    lam = Const.h * Const.c / E
    wavelength = lam / Const.A
    return wavelength


def photon_energy(wavelength_a):
    """
    Converts wavelength in A to energy in keV
     energy_kev = photon_energy(wavelength_a)
     Energy [keV] = h*c/L = 12.3984 / lambda [A]
    """

    # SI: E = hc/lambda
    lam = wavelength_a * Const.A
    E = Const.h * Const.c / lam

    # Electron Volts:
    energy = E / Const.e
    return energy / 1000.0


def wavevector(wavelength_a):
    """Return wavevector = 2pi/lambda"""
    return 2 * np.pi / wavelength_a


def bmatrix(a, b=None, c=None, alpha=90., beta=90., gamma=90.):
    """
    Calculate the B matrix as defined in Busing&Levy Acta Cyst. 22, 457 (1967)
    Creates a matrix to transform (hkl) into a cartesian basis:
        (qx,qy,qz)' = B.(h,k,l)'       (where ' indicates a column vector)

    The B matrix is related to the reciprocal basis vectors:
        (astar, bstar, cstar) = 2 * np.pi * B.T
    Where cstar is defined along the z axis

    The B matrix is related to the real-space unit vectors:
        (A, B, C) = B^-1 = inv(B)

    :param a: lattice parameter a in Anstroms
    :param b: lattice parameter b in Anstroms
    :param c: lattice parameter c in Anstroms
    :param alpha: lattice angle alpha in degrees
    :param beta: lattice angle beta in degrees
    :param gamma: lattice angle gamma in degrees
    :returns: [3x3] array B matrix in inverse-Angstroms (no 2pi)
    """
    if b is None:
        b = a
    if c is None:
        c = a
    alpha1 = np.deg2rad(alpha)
    alpha2 = np.deg2rad(beta)
    alpha3 = np.deg2rad(gamma)

    beta1 = np.arccos((np.cos(alpha2) * np.cos(alpha3) - np.cos(alpha1)) / (np.sin(alpha2) * np.sin(alpha3)))
    beta2 = np.arccos((np.cos(alpha1) * np.cos(alpha3) - np.cos(alpha2)) / (np.sin(alpha1) * np.sin(alpha3)))
    beta3 = np.arccos((np.cos(alpha1) * np.cos(alpha2) - np.cos(alpha3)) / (np.sin(alpha1) * np.sin(alpha2)))

    b1 = 1 / (a * np.sin(alpha2) * np.sin(beta3))
    b2 = 1 / (b * np.sin(alpha3) * np.sin(beta1))
    b3 = 1 / (c * np.sin(alpha1) * np.sin(beta2))

    c1 = b1 * b2 * np.cos(beta3)
    c2 = b1 * b3 * np.cos(beta2)
    c3 = b2 * b3 * np.cos(beta1)

    bm = np.array([
        [b1, b2 * np.cos(beta3), b3 * np.cos(beta2)],
        [0, b2 * np.sin(beta3), -b3 * np.sin(beta2) * np.cos(alpha1)],
        [0, 0, 1 / c]
    ])
    return bm


import numpy as np


def norm_vector(vector, min_mag=0.001):
    mag = np.linalg.norm(vector)
    if mag < min_mag:
        mag = 1.
    return np.divide(vector, mag)


def rot_matrix(angle_rad: float, axis=(0, 0, 1)):
    """
    Generate rotation matrix about arbitary axis
    https://en.wikipedia.org/wiki/Rotation_matrix#Rotation_matrix_from_axis_and_angle
    """
    axis = norm_vector(axis)
    ux, uy, uz = axis
    c = np.cos(angle_rad)
    s = np.sin(angle_rad)
    c1 = 1 - c
    r = np.array([
        [
            (ux * ux * c1) + c,
            (uy * ux * c1) - uz * s,
            (uz * ux * c1) + uy * s
        ],
        [
            (ux * uy * c1) + uz * s,
            (uy * uy * c1) + c,
            (uz * uy * c1) - ux * s,
        ],
        [
            (ux * uz * c1) - uy * s,
            (uy * uz * c1) + ux * s,
            (uz * uz * c1) + c,
        ]
    ])
    return r


def rotation_t_matrix(value=0.0, vector=(0, 0, 1), offset=(0, 0, 0)):
    """
    Create 4x4 transformation matrix including a rotation
    """
    t = np.eye(4)
    t[:3, :3] = rot_matrix(angle_rad=value, axis=vector)
    t[:3, 3] = offset
    return t


def translation_t_matrix(value=0.0, vector=(0, 0, 1), offset=(0, 0, 0)):
    """
    Create 4x4 transformation matrix including a translation
    """
    t = np.eye(4)
    translation = value * np.reshape(vector, 3) + np.reshape(offset, 3)
    t[:3, 3] = translation
    return t


def rotate_by_matrix(xyz, angle_deg=0.0, axis=(0, 0, 1)):
    r = rot_matrix(np.deg2rad(angle_deg), axis)
    xyz = np.reshape(xyz, (-1, 3))
    return np.dot(r, xyz.T).T


def transform_by_t_matrix(xyz, t_matrix):
    xyz = np.reshape(xyz, (-1, 3))
    return (np.dot(t_matrix[:3, :3], xyz.T) + t_matrix[:3, 3:]).T
